"""Package for calculating Similarity-sensitive diversities.

The diversity measures that can be calculated are described rigorously
in https://arxiv.org/abs/1404.6520.

See greylock.metacommunity.Metacommunity
"""

from greylock.metacommunity import Metacommunity
